/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along
 with this program; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */


#include "Editors/ScoreEditor.h"
#include "Midi/Sequence.h"
#include "Midi/Track.h"
#include "Pickers/VolumeSlider.h"
#include "Pickers/TuningPicker.h"
#include "Images/ImageProvider.h"
#include "Images/Drawable.h"
#include "GUI/GLPane.h"

#include "OpenGL.h"

#include <string>

#include "Editors/RelativeXCoord.h"

namespace AriaMaestosa {
	
const int y_dist = 8;
const int first_line_position = y_dist * 5;
const int line_amount = 5;
	
ScoreEditor::ScoreEditor(Track* track) : Editor(track)
{
	updateNoteDestChart();
}

enum tail
{
	TAIL_UP,
	TAIL_DOWN,
	TAIL_NONE
};

enum sign
{
	SIGN_SHARP,
	SIGN_FLAT,
	SIGN_NATURAL,
	SIGN_NONE
};

void ScoreEditor::drawNote(int note_x, int string, bool open, int tail_type, int tail_amount, bool pointed, int sign)
{
	const int note_y_in_pixels = getEditorYStart() + first_line_position + (y_dist/2)*(string-1) - ImageProvider::getImages()->noteOpenImg->height/2;
	
	// main round
	if(open)
	{
		glEnable(GL_TEXTURE_2D);
		noteOpen->move(getEditorXStart() + note_x, note_y_in_pixels);
		noteOpen->render();
		glDisable(GL_TEXTURE_2D);
		glLoadIdentity();
	}
	else
	{
		glEnable(GL_TEXTURE_2D);
		noteClosed->move(getEditorXStart() + note_x, note_y_in_pixels);
		noteClosed->render();
		glDisable(GL_TEXTURE_2D);
		glLoadIdentity();
	}
	
	// pointed
	if(pointed)
	{
		glBegin(GL_POINTS);
		glVertex2f(getEditorXStart() + note_x + 14, note_y_in_pixels + 5);
		glEnd();
	}
	
	// tail type
	if(tail_type == TAIL_UP)
	{
		glLoadIdentity();
		glBegin(GL_LINES);
		glVertex2f( getEditorXStart() + note_x + 9, getEditorYStart() + first_line_position + (y_dist/2)*(string-1) );
		glVertex2f( getEditorXStart() + note_x + 9, getEditorYStart() + first_line_position + (y_dist/2)*(string-1) - 30 );
		glEnd();
	}
	else if(tail_type == TAIL_DOWN)
	{
		glLoadIdentity();
		glBegin(GL_LINES);
		glVertex2f( getEditorXStart() + note_x + 9, getEditorYStart() + first_line_position + (y_dist/2)*(string-1) );
		glVertex2f( getEditorXStart() + note_x + 9, getEditorYStart() + first_line_position + (y_dist/2)*(string-1) + 30 );
		glEnd();
	}
}

/*
 * This is called when user changes the tuning. This tells the editor to change notes so that they match the new settings.
 */


void ScoreEditor::render(int dragX_arg, int mousey_current, int XBeforeDrag_arg, int mousey_initial, bool focus)
{
    
    if(!imagesLoaded()) return;
    
    // prepare coords
    RelativeXCoord<int> mousex_current(sequence);
    mousex_current.setValue(dragX_arg, WINDOW);
    
    RelativeXCoord<int> mousex_initial(sequence);
    mousex_initial.setValue(XBeforeDrag_arg, WINDOW);
    
    glEnable(GL_SCISSOR_TEST);
    // glScissor doesn't seem to follow the coordinate system so this ends up in all kinds of weird code to map to my coord system (from_y going down)
    glScissor(10, sequence->glPane-> getHeight() - (20+height + from_y+barHeight+20), width-15, 20+height);
    
	
	
    // white background
    glDisable(GL_TEXTURE_2D);
    
    glColor3f(1,1,1);
    glBegin(GL_QUADS);
    
    glVertex2f( 0, getEditorYStart());
    glVertex2f( 0, getYEnd());
    glVertex2f( getXEnd(), getYEnd());
    glVertex2f( getXEnd(), getEditorYStart());
    
    glEnd();
    
    drawVerticalMeasureLines(getEditorYStart()+first_line_position, getEditorYStart()+first_line_position+line_amount*y_dist-y_dist);
	
	// --------------------------- grey left part -----------------------------
    glLoadIdentity();
    
    // grey background
    if(!focus) glColor3f(0.4, 0.4, 0.4);
    else glColor3f(0.8, 0.8, 0.8);
    glBegin(GL_QUADS);
    
    glVertex2f( 0, getEditorYStart());
    glVertex2f( 0, getYEnd());
    glVertex2f( getEditorXStart()-3, getYEnd());
    glVertex2f( getEditorXStart()-3, getEditorYStart());
    
    glEnd();
	
	glEnable(GL_TEXTURE_2D);
	keyG->move(getEditorXStart() - 65, getEditorYStart() + first_line_position-11);
	keyG->render();
	glDisable(GL_TEXTURE_2D);
	
    // ------------------------------- draw lines -------------------------------
    glColor3f(0,0,0);
	glLoadIdentity();
    glBegin(GL_LINES);
    
	
    for(int n=0; n<line_amount; n++)
	{
        glVertex2f( 0, getEditorYStart()+first_line_position+n*y_dist );
        glVertex2f( getXEnd(), getEditorYStart()+first_line_position+n*y_dist );
    }
    
    glEnd();

    
    // ---------------------- draw notes ----------------------------
	
	glColor3f(0,0,0);
	glLineWidth(2);
	glPointSize(4);
	
	//drawNote(25, 3, true, TAIL_UP, 0, false, SIGN_NONE);
	//drawNote(50, 4, false, TAIL_DOWN, 0, true, SIGN_SHARP);
	//drawNote(75, 5, true, TAIL_NONE, 0, false, SIGN_FLAT);
	//drawNote(int note_x, int string, bool open, int tail_type, int tail_amount, bool pointed, int sign)
		
	//noteClosed->move(getEditorXStart(), getEditorYStart() + first_line_position + y_dist*3 - ImageProvider::getImages()->noteOpenImg->height/2);
	//noteClosed->render();
	
	const int measureLength = track->sequence->ticksPerMeasure();
    const int noteAmount = track->getNoteAmount();
    for(int n=0; n<noteAmount; n++)
	{
      // ...
		const int notePitch = track->getNotePitchID(n);
		const int noteLine = noteDestChart[notePitch].line;
		const float noteLength =  (float)(track->getNoteEndInMidiTicks(n) - track->getNoteStartInMidiTicks(n)) / (float)measureLength;
		if(noteLine != -1)
		{
			int tail = (noteLine>=5 ? TAIL_UP : TAIL_DOWN);
			if(noteLength==1) tail=TAIL_NONE; // whole note
			const bool filled = !(noteLength < 0.5); // notes smaller than halves are filled
			
			drawNote(track->getNoteStartInPixels(n), noteLine, filled, tail, 0, false, noteDestChart[notePitch].sign);
		}
    }//next
    
    
       
    glLineWidth(1);
    
    
    glEnable(GL_TEXTURE_2D);
    glDisable(GL_SCISSOR_TEST);
    
}

void ScoreEditor::mouseDown(const int x_arg, const int y)
{

    // prepare coords
    RelativeXCoord<int> x(sequence);
    x.setValue(x_arg, WINDOW);
        
}
void ScoreEditor::mouseDrag(const int mousex_current, const int mousey_current, const int mousex_initial, const int mousey_initial)
{
    
    
}
void ScoreEditor::mouseUp(const int dragX_arg, const int mousey_current, const int XBeforeDrag_arg, const int mousey_initial)
{

    // prepare coords
    RelativeXCoord<int> mousex_current(sequence);
    mousex_current.setValue(dragX_arg, WINDOW);
    
    RelativeXCoord<int> mousex_initial(sequence);
    mousex_initial.setValue(XBeforeDrag_arg, WINDOW);
     
}

void ScoreEditor::rightClick(const int x, const int y)
{
        
    
}

int ScoreEditor::getYScrollInPixels()
{
    /*Unused*/
    return -1;
}

ScoreEditor::~ScoreEditor()
{
}

void ScoreEditor::updateNoteDestChart()
{
	// middle G = 64
	
	/*
	 Below, 'note' stands for:
	 B  : 0
	 A# : 1
	 A  : 2
	 G# : 3
	 G  : 4
	 F# : 5
	 F  : 6
	 E  : 7
	 D# : 8
	 D  : 9
	 C# : 10
	 C  : 11
	*/
	
	int map_note_to_line[12] = {0, -1, 1, -1, 2, -1, 3, 4, -1, 5, -1, 6};
	
	
	for(int n=0; n<128; n++)
	{
		//const int suggested_line = (n-64)+9;
		const int octave = n/12;
		const int note = n%12;
		
		int suggested_line = -1;
		if(map_note_to_line[note] != -1) suggested_line = (octave-5)*7 + map_note_to_line[note] + 5;
		
		if(suggested_line>-10 and suggested_line<=24)
		{
			noteDestChart[n].line = suggested_line;
			noteDestChart[n].sign = SIGN_NONE;
			//std::cout << n << " : " << suggested_line << std::endl;
		}
		else
			// out of the range of the score
		{
			noteDestChart[n].line = -1;
			noteDestChart[n].sign = SIGN_NONE;
		}
	}
}

}
