#include "IO/NotationExport.h"
#include "Midi/Sequence.h"

#include <wx/wx.h>
#include <wx/sizer.h>
#include <wx/file.h>

#include "LeakCheck.h"
#include "GUI/GraphicalTrack.h"
#include "Editors/GuitarEditor.h"
#include "IO/IOUtils.h"

namespace AriaMaestosa
{
	
	enum IDS
{
	ID_OK,
	ID_CANCEL
};

void completeExport(bool accepted);
void exportTablature(Track* t, wxFile* file);

bool ignoreMuted_bool = false;
bool ignoreHidden_bool = false;
bool checkRepetitions_bool = false;

// ----------------------------------------------------------------------------------------------------
// ------------------------------------------- setup dialog -------------------------------------------
// ----------------------------------------------------------------------------------------------------

class NotationSetup : public wxDialog
{
	wxCheckBox* ignoreHidden;
	wxCheckBox* ignoreMuted;
	wxCheckBox* detectRepetitions;
	
	wxPanel* buttonPanel;
	wxButton* okButton;
	wxButton* cancelButton;
	
	wxBoxSizer* boxSizer;
	
	int dlg_id;
	
public:
		
		DECLARE_LEAK_CHECK();
	
	NotationSetup(int mode = -1) : wxDialog(NULL, wxID_ANY, wxGetTranslation("Export to musical notation"), wxPoint(200,200), wxSize(200,400))
    {
		INIT_LEAK_CHECK();
		
		boxSizer=new wxBoxSizer(wxVERTICAL);
		
		if(mode==-1)
		{
			// "ignore hidden tracks" checkbox
			ignoreHidden=new wxCheckBox(this, wxID_ANY, wxGetTranslation("Ignore hidden tracks"));
			ignoreHidden->SetValue(true);
			boxSizer->Add(ignoreHidden, 1, wxALL, 5);
			
			// "ignore muted tracks" checkbox
			ignoreMuted=new wxCheckBox(this, wxID_ANY, wxGetTranslation("Ignore muted tracks"));
			ignoreMuted->SetValue(true);
			boxSizer->Add(ignoreMuted, 1, wxALL, 5);
		}
		else
		{
			ignoreHidden = NULL;
			ignoreMuted = NULL;
		}
		
		// "Show repeated measures only once" checkbox
		detectRepetitions=new wxCheckBox(this, wxID_ANY, wxGetTranslation("Show repeated measures (e.g. chorus) only once"));
		detectRepetitions->SetValue(true);
		boxSizer->Add(detectRepetitions, 1, wxALL, 5);
		
		// OK-Cancel buttons
		{
			buttonPanel = new wxPanel(this);
			boxSizer->Add(buttonPanel, 0, wxALL, 0);
			
			wxBoxSizer* subsizer = new wxBoxSizer(wxHORIZONTAL);
			
			okButton=new wxButton(buttonPanel, ID_OK, wxT("OK"));
			okButton->SetDefault();
			subsizer->Add(okButton, 0, wxALL, 15);
			
			cancelButton=new wxButton(buttonPanel, ID_CANCEL, wxGetTranslation("Cancel"));
			subsizer->Add(cancelButton, 0, wxALL, 15);
			
			buttonPanel->SetSizer(subsizer);
			buttonPanel->SetAutoLayout(true);
			subsizer->Layout();
		}
		
		SetAutoLayout(true);
		SetSizer(boxSizer);
		boxSizer->Layout();
		boxSizer->SetSizeHints(this);
		
		Center();
		dlg_id = ShowModal();

    }
	
	void cancelClicked(wxCommandEvent& evt)
	{
		EndModal(dlg_id);
		completeExport(false);
	}
	
	void okClicked(wxCommandEvent& evt)
	{
		if(ignoreMuted != NULL)
		{
			ignoreMuted_bool = ignoreMuted->IsChecked();
			ignoreHidden_bool = ignoreHidden->IsChecked();
		}
		checkRepetitions_bool = detectRepetitions->IsChecked();
		EndModal(dlg_id);
		completeExport(true);
	}
	
	DECLARE_EVENT_TABLE();
	
};

BEGIN_EVENT_TABLE(NotationSetup, wxDialog)

EVT_BUTTON(ID_OK, NotationSetup::okClicked)
EVT_BUTTON(ID_CANCEL, NotationSetup::cancelClicked)

// FIXME - handle when window closes - leak!

END_EVENT_TABLE()

NotationSetup* setup;
Sequence* currentSequence;
Track* currentTrack;

// ----------------------------------------------------------------------------------------------------
// ------------------------------------- first function called ----------------------------------------
// ----------------------------------------------------------------------------------------------------

// user wants to export to notation - remember what is the sequence, then show set-up dialog
void exportNotation(Sequence* sequence)
{	
	currentSequence = sequence;
	currentTrack = NULL;
	setup = new NotationSetup(GUITAR);
}

void exportNotation(Track* t)
{	
	currentTrack = t;
	currentSequence = NULL;
	setup = new NotationSetup(GUITAR);
}

// ----------------------------------------------------------------------------------------------------
// ---------------------------------------- main writing func -----------------------------------------
// ----------------------------------------------------------------------------------------------------
// after dialog is shown and user clicked 'OK' this is called to complete the export
void completeExport(bool accepted)
{
	delete setup;	
	if(!accepted) return;
	
show_dialog:
		
		if(currentSequence == NULL) currentSequence = currentTrack->sequence;
	
		// ask user to select file destination
		wxString filepath = wxFileSelector(wxGetTranslation("Select destination file"),
										   wxT(""),
										   wxString(currentSequence->sequenceName)+wxString(".txt"),
										   wxT(".txt"),
										   wxT("text file|*.txt"), wxSAVE);
	
	if(filepath.IsEmpty()) return; // user cancelled
	
	if(!filepath.IsAscii())
	{
		wxMessageBox(wxGetTranslation("Invalid file name. (Sorry, accentuated characters are not supported)"));	
		goto show_dialog;
	}
	
	
	if( wxFileExists(filepath) )
	{
		int answer = wxMessageBox( wxGetTranslation("The file already exists. Do you wish to overwrite it?"), wxGetTranslation("Confirm"),
								   wxYES_NO);
		if (answer != wxYES) return;
	}
	
	
	// open file and prepare for writing to it
	wxFile* file = new wxFile( filepath, wxFile::write );
	
	// we want to export the entire song
	if(currentTrack == NULL)
	{
		
		// iterate through all the the tracks of the sequence, only consider those that are visible and not muted
		const int track_amount = currentSequence->getTrackAmount();
		for(int n=0; n<track_amount; n++)
		{
			Track* track = currentSequence->getTrack(n);
			if( (track->graphics->muted and ignoreMuted_bool) or
				((track->graphics->collapsed or track->graphics->docked) and ignoreHidden_bool)
				)
				// track is disabled, ignore it
				continue;
			
			std::cout << "Generating notation for " << track->getName() << std::endl;
			const int mode = track->graphics->editorMode;
			
			if( mode == GUITAR)
			{
				
				exportTablature( track, file );
			}
			else
			{
				std::cout << "unsupported view"	<< std::endl;
			}

		}// next track

	}
	// we want to export a single track
	else
	{
		const int mode = currentTrack->graphics->editorMode;
		
		if( mode == GUITAR)
		{
			exportTablature( currentTrack, file );
		}
		else
		{
			std::cout << "unsupported view"	<< std::endl;
		}
		
	}
	
	file->Flush();
	file->Close();
	delete file;
}

// ----------------------------------------------------------------------------------------------------
// ---------------------------------------- tab writing func ------------------------------------------
// ----------------------------------------------------------------------------------------------------
void exportTablature(Track* track, wxFile* file)
{
	//file->Write("W whole\nH half\nQ quarter\nE 8th\nS 16th\nT 32th\nX 64th\n\n");
	
	const int ticksPerMeasure = currentSequence->ticksPerMeasure();
	const int measureAmount = currentSequence->getMeasureAmount();
	const int ticksPerBeat = currentSequence->ticksPerBeat();
	const int noteAmount = track->getNoteAmount();
	GuitarEditor* editor = track->graphics->guitarEditor;
	const int string_amount = editor->tuning.size();
	
	const int max_length_of_a_line = 110;
	
	int note=0;
	
	// init one string for each string
	int measures_appended = 0;
	wxString strings[ string_amount ];
	wxString title_line;
	
	for(int s=0; s<string_amount; s++)
	{
		strings[s] = wxT("|");	
	}
	title_line = wxT("");
	
	int firstTickInMeasure[measureAmount];
	int lastTickInMeasure[measureAmount];
	int measureFirstNote[measureAmount];
	int measureLastNote[measureAmount];
	
	for(int measure=0; measure<measureAmount; measure++)
	{
		
		int zoom = 1;
		
		int added_chars = 0;
		
		// fret number higher than 9 take more place, so we one is met, aditionnal room is added
		// however, if 2 of these play at the same time, we don't need twice more space. so this bool
		// remembers whether we already met one of these fret numbers
		bool found_fret_higher_than_9=false;
		
		firstTickInMeasure[measure] = measure * ticksPerMeasure;
		lastTickInMeasure[measure] = (measure + 1) * ticksPerMeasure; // where does the current measure end
		
		
		// first note in measure (which is also last note of previous measure, that was set in last iteration of the for loop)
		measureFirstNote[measure] = note;
		
		// ------------------------------------ find what the first, last and shortest note in current measure --------------------------------------
		// find what is the shortest note in this measure. this will help determine how big it must be made.
		// to do this, iterate through notes, keeping the smallest length we find, stopping when we reach next measure
		// this by the way finds where the measure begins and end, thus filling up 'measureFirstNote' and 'measureLastNote' variables.
		int shortestDuration = -1;
		for(; note<noteAmount; note++)
		{
			// stop when we're at next measure - it will be done in next measure iteration
			if( track->getNoteStartInMidiTicks(note) >= lastTickInMeasure[measure] ) break;
			
			const int currentNoteDuration = track->getNoteEndInMidiTicks(note) - track->getNoteStartInMidiTicks(note);
			if( currentNoteDuration < shortestDuration or shortestDuration==-1) shortestDuration = currentNoteDuration;
			
		}
		
		measureLastNote[measure] = note; // last note in measure
		
		// if empty measure
		if(shortestDuration==-1)
		{
			for(int s=0; s<string_amount; s++)
			{
				strings[s].Append("--|");
			}
			title_line.Append("   ");
			continue;
		} 
		
		// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
		// if user checked "detect repetitions"
		if(checkRepetitions_bool)
		{
			bool found_repetition = false;
			
			// check current measure against all previous measures to see if it is not a repetition
			for(int checkMeasure=0; checkMeasure<measure; checkMeasure++)
			{
				//std::cout << "i'm at measure " << (measure+1) << ", let's check measure " << (checkMeasure+1) << std::endl;
				
				// if these 2 measures don't even have the same number of notes, they're definitely not the same
				if(
				   (measureLastNote[checkMeasure] - measureFirstNote[checkMeasure]) !=
				   (measureLastNote[measure] - measureFirstNote[measure])
				   )
				{
					//std::cout << "** they don't have the same amount of notes (" << measureLastNote[checkMeasure] << "-" << measureFirstNote[checkMeasure] <<
					//" != " << measureLastNote[measure] << "-" << measureFirstNote[measure] << ")" << std::endl;
					continue;
				}
				
				// if we get till there, the 2 measures have the same amount of notes.
				// to know whether they are truly identitcal, we need to compare note by note
				bool dont_match=false;
				for(int checkNote=0; checkNote<(measureLastNote[checkMeasure] - measureFirstNote[checkMeasure]); checkNote++)
				{
					// check start tick matches
					if(track->getNoteStartInMidiTicks(measureFirstNote[checkMeasure] + checkNote) - firstTickInMeasure[checkMeasure] !=
					   track->getNoteStartInMidiTicks(measureFirstNote[measure] + checkNote) - firstTickInMeasure[measure]
					   )
					{
						//std::cout << "** note #" << checkNote << " don't have same start" << std::endl;
						dont_match = true;
						break; // this measure doesn't match, stop checking it
					}
					
					// check end tick matches
					if(track->getNoteEndInMidiTicks(measureFirstNote[checkMeasure] + checkNote) - firstTickInMeasure[checkMeasure] !=
					   track->getNoteEndInMidiTicks(measureFirstNote[measure] + checkNote) - firstTickInMeasure[measure]
					   )
					{
						//std::cout << "** note #" << checkNote << " don't have same end" << std::endl;
						dont_match = true;
						break; // this measure doesn't match, stop checking it
					}
					
					// check pitch matches
					if(track->getNotePitchID(measureFirstNote[checkMeasure] + checkNote) !=
					   track->getNotePitchID(measureFirstNote[measure] + checkNote)
					   )
					{
						//std::cout << "** note #" << checkNote << " don't have same pitch" << std::endl;
						dont_match = true;
						break; // this measure doesn't match, stop checking it
					}
					
				}//next
				
				if(dont_match) continue; // try another measure
				
				// if we get there, the 2 measures fully match!
				const wxString repeatMeasureName = to_wxString(checkMeasure+1);
				const int chars_in_repeat_measure_name = repeatMeasureName.Length();
				for(int n=0; n<string_amount; n++)
				{
					if(n == string_amount/2) strings[n].Append( wxT(" *") + repeatMeasureName + wxT(" |") );
					else strings[n].Append('-', chars_in_repeat_measure_name+3 ).Append("|");
				}
				title_line.Append(' ', chars_in_repeat_measure_name+4 );
					
				found_repetition = true;
				break;
				
			}//next
			
			if(found_repetition) continue; // jump to next measure
		}//endif
		// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
		
		// we found the length of the shortest note in this measure
		const int divider = (int)(
								  currentSequence->getMeasureTopValue() * (float)ticksPerBeat / (float)shortestDuration
								  );
		
		// if notes are very long, zoom a bit because we don't want a too short measure
		if( divider <= 2 ) zoom = 4;
		if( divider <= 1 ) zoom = 8;
		
		// contains the number of characters each line contained, before adding new ones for this measure.
		int charAmountWhenBeginningMeasure;
		
		// ------------------------------------ check for end of line --------------------------------------
		// if current line is full (i.e. no more horizontal space), output current line to file and prepare a new line
		
		// check how much place the next measure would take
		const int needed_length = (int)(
										divider * currentSequence->getMeasureTopValue()  / 2.0
										);
		
		// if it takes too much space, end the current line and begin a new one
		if( measures_appended > 0 and (int)strings[0].Length() + needed_length > max_length_of_a_line)
		{
			file->Write( title_line + wxT("\n") );
			for(int s=0; s<string_amount; s++)
			{
				file->Write( strings[s] + wxT("\n"));
				// for last string, skip one more line
				strings[s] = wxT("|");	;
			}
			file->Write( wxT("\n") );
			measures_appended = 0;
			charAmountWhenBeginningMeasure = 2;
			title_line = wxT("");
		}
		
		// ------------------------------------ begin the new measure --------------------------------------
		for(int s=0; s<string_amount; s++)
		{
			strings[s].Append("-");
		}
		charAmountWhenBeginningMeasure = strings[0].Length();
		
		// ------------------------------------ write notes on the strings --------------------------------------
		int currentStep = 0;
		// use the IDs of the first and the last note in the measure we previously found, and iterate through all the notes in the measure
		for(int n=measureFirstNote[measure]; n<measureLastNote[measure]; n++)
		{
			// gather information about current note
			const int note_fret = track->getNoteFret(n);
			const int note_string = track->getNoteString(n);
			
			// here, what i call "step" is a "-" followed by a number or another "-". e.g. tab |-2-3-4---5-| contains 5 steps.
			// think of it as X Position, but counted in characters
			const int note_step = (int)(
										(float)divider * zoom *
										(float)(track->getNoteStartInMidiTicks(n) - (firstTickInMeasure[measure]) ) /
										(float)ticksPerMeasure
										);
			
			// if current note is farther than current step, add "-"s until we reach the position of the current's note step
			if(note_step > currentStep)
			{
				// we're moving to a new lcoation, so reset this variable to be able to find more 2-digit fret numbers
				found_fret_higher_than_9 = false;
				
				// calculate how many characters we need on each string to go to next note
				const int wantedStep = currentStep + (note_step - currentStep);
				
				const int wantedChars = charAmountWhenBeginningMeasure + wantedStep*2 + added_chars;
				for(int s=0; s<string_amount; s++)
				{
					while( (int)strings[s].Length() < wantedChars ) strings[s].Append("-");
				}
				currentStep = wantedStep;
			}
			
			if(note_fret > 9 and !found_fret_higher_than_9)
			{
				added_chars++;
				found_fret_higher_than_9 = true;
			}
			strings[note_string].Append( to_wxString(note_fret)  );
		}
		
		// ------------------------------------ end of measure --------------------------------------
		// fill all lines that are not already full with "-"s
		const int wantedChars = charAmountWhenBeginningMeasure + divider * 2 * zoom + added_chars;
		for(int s=0; s<string_amount; s++)
		{
			while( (int)strings[s].Length() < wantedChars) strings[s].Append("-");
		}
		
		// ------------------------------------ title line --------------------------------------
		// create title line for this measure
		wxString measureTitle = to_wxString(measure+1)/* + wxT(" ")*/;
		title_line.Append( measureTitle );
		const int title_length = measureTitle.Length()/* + 1 */;
		//const int beatAmount = currentSequence->getMeasureTopValue();
		const int chars_in_measure = (divider * 2 * zoom + added_chars);
		//const int charsPerBeat = (int)round( (float)chars_in_measure / (float)beatAmount );
		
		for(int i=title_length; i<chars_in_measure; i++)
		{/*
		 if( charsPerBeat != 0 and
			 (i-1 - title_length - 3*(divider/8)) % charsPerBeat == 0 ) title_line.Append(".");	
		 else */title_line.Append(" ");	
		}
		while( (int)title_line.Length() < wantedChars ) title_line.Append(" ");
		
		currentStep = 0;
		
		// end measure with a vertical bar
		for(int s=0; s<string_amount; s++)
		{
			strings[s].Append("|");
		}
		
		measures_appended++;
		
	} // next measure
	
	// output to file anything that remains
	if( measures_appended > 0)
	{
		file->Write( title_line + wxT("\n") );
		for(int s=0; s<string_amount; s++)
		{
			file->Write( strings[s] + wxT("-|\n") );
		}
		measures_appended = 0;
	}
	
}

}
